############################################################################
# Matt Rodriguez, LBNL
#Copyright (c) 2003, The Regents of the University of California,
#through Lawrence Berkeley National Laboratory 
#(subject to receipt of any required approvals from the U.S. Dept. of Energy). 
#All rights reserved. 
###########################################################################
"""
Provides an interface to the Gass Transfer module.
"""

from pyGlobus import utilc
from pyGlobus.util import GlobusException
import gassTransferc
from threading import *

class TransferException(GlobusException):
    pass

class TransferServerException(GlobusException):
    pass

class TransferReqAttrException(GlobusException):
    pass

class TransferListenerAttrException(GlobusException):
    pass

class TransferListenerException(GlobusException):
    pass

class GassTransferServerException(GlobusException):
    pass

class TransRequest:
    """
    Simple class that wraps the request handle for clients and server
    handle is the request handle pointer
    isClient is true if it is a client-side handle

    """
    def __init__(self,handle, isClient):
        self._handle = handle
        self._isClient = isClient
        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferException("Unable to initialize the module")
            raise ex
        self._activated = 1 
        return

    def __del__(self):
        ret = utilc.globus_module_deactivate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferException("Unable to deactivate the module")
            print ex
        ret = gassTransferc.request_destroy(self._handle)
        if not ret == 0:
            ex = TransferException("Can't destroy the request_handle")
            print ex
        return
   
    def get_handle(self):
        return self._handle


    def get_request_type(self):
        """Returns a string the request type of the Request handle"""
        reqType = gassTransferc.get_request_type(self._handle)
        if reqType == 0:
            return "INVALID"
        elif reqType == 1:
            return "GET"
        elif reqType == 2:
            return "PUT"
        elif reqType == 3:
            return "APPEND"
        else:
            raise RuntimeError, "Invalid request for TransData object, aborting!"
    
    
    def get_url(self):
        """Returns the url associated with the request """
        ret = gassTransferc.get_url(self._handle)
        return ret

    def get_subject(self):
        """Returns the subject associated with the request, the subject has to be set by the user,
              Setting the subject of a request is a feature that has not been implemented yet
              """
        ret = gassTransferc.get_subject(self._handle)
        return ret

    def get_status(self):
        """Return the status of the request handle"""
        ret = gassTransferc.get_request_status(self._handle)
        return ret
    

   
class TransListener:
    """ Simple class that wraps the listener handle for servers
       handle is the listerner handle pointer

       The TransServer class provides an higher level interface than this.
       
       """
    
    def __init__(self, attr=None, scheme= "http"):
        """
        **Arguments**
        
        *attr* is a ListenerAttr object
        
        *scheme* is a string denoting the scheme used for file transfer eg.: http
        
        """

        if attr == None:
            attr = ListenerAttr(scheme)
            
        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        
        
        if not ret == 0:
            ex = TransferReqAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
    
        ret,listener = gassTransferc.create_listener(attr.get_handle(), scheme)
        if not ret == 0:
            ex = TransferServerException(listener)
            raise ex

        self._handle= listener
   
     
     
    def __del__(self):
        """
        frees the request handles memory
        
        *Raises*
        A TransferException is raised if there is a problem freeing the memory
        """
        gassTransferc.listener_destroy(self._handle)
        utilc.globus_module_deactivate(gassTransferc.get_module())
        

    def close_listener(self,callback, arg):
        """
	Callbacks for any outstanding register_accepts are called before the listener is closed
	**Arguements**

	- *callback* function that is called when the listener is closed
	
	- *arg* user supplied arguement to the callback function
	
	**Throws**
	A TransferListenerException if the listener cannot be closed
	"""
        
        ret, cbHandle, listener = gassTransferc.close_listener(self._handle, callback, arg)
        if not ret == 0:
            ex = TransferListenerException(cbHandle)
            raise ex

        return cbHandle
     
    def get_handle(self):
        return self._handle

    def get_base_url(self):
        """
	Gets the base url that the listener is listening on
        
	Returns a string of that url
             """
        url = gassTransferc.get_base_url(self._handle)
        return url


class TransferReqAttr:

    def __init__(self, urlScheme, handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        
        if not ret == 0:
            ex = TransferReqAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        
        if handle == None:
            ret, handle = gassTransferc.reqattr_init(urlScheme)
        if not ret == 0:
            ex = TransferReqAttrException("unable to init the attr")
            raise ex
        self._handle = handle
        return

    def __del__(self):
        ret = gassTransferc.reqattr_destroy(self._handle)
        if not ret == 0:
            ex = TransferReqAttrException("unable to destroy the attr")
            print ex        
        ret = utilc.globus_module_deactivate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferReqAttrException("Unable to deactivate the module")
            print ex
        return

    ########################################################################
    # XXX This handle is only valid as long as this object exists.  When
    # the object is destroyed, the memory this handle points to is freed.
    #######################################################################
    def get_handle(self):
        return self._handle


class ListenerAttr:

    def __init__(self, urlScheme= "http", handle = None):
        self._activated = 0
        ret = utilc.globus_module_activate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferListenerAttrException("Unable to initialize the module")
            raise ex
        self._activated = 1
        if handle == None:
            ret, handle = gassTransferc.listenerattr_init(urlScheme)
        if not ret == 0:
            ex = TransferListenerAttrException("unable to init the listenattr")
            raise ex
        self._handle = handle
        return

    def __del__(self):
        ret = gassTransferc.listenerattr_destroy(self._handle)
        if not ret == 0:
            ex = TransferListenerAttrException("Unable to destroy listenerAttr")
            print ex
        ret = utilc.globus_module_deactivate(gassTransferc.get_module())
        if not ret == 0:
            ex = TransferReqAttrException("Unable to deactivate the module")
            print ex
        return
        
    def get_handle(self):
        return self._handle

class TransData:
    """This class handles send_bytes and receive_bytes calls which both a TransferClient and a TransferServer
       object can make. This class is the parent class for those two classes.
       """
    def __init__(self):
        
        self._cbList = []
        return

    def __del__(self):
        
        
        for cb in self._cbList:
            self.free_callback(cb)
            
        return

    def free_callback(self, cbHandle):
        """
               Free a callback.

               Free's the underlying memory allocated for a callback structure.

               **Arguments:**

               - *cbHandle* A SWIG'ized pointer to a callback structure.
               """
        gassTransferc.free_callback(cbHandle)
        indx = self._cbList.index(cbHandle)
        del self._cbList[indx]
        return

    def send_bytes(self, request, buffer,eof,callback, arg):
        """
              After authorizing a request the server calls send bytes which sends the data in the buffer object

              **Arguements**

              -*request* a server-side request if this is in conjunction with a get request,
              if in conjunction with a put or append this should be a client-side request
        
              - *buffer* A pyGlobus.util.Buffer.Buffer object.. This hides the
              internal globus_byte_t data buffer.

              - *eof* true if this is the last data to be sent, false otherwise

              - *callback* is a function that is called after the data in the buffer has been sent

              - *arg* a user supplied argument to the callback function  

              **Returns:**

              A SWIG'ized pointer to the callback handle is returned. After the callback
              completes, this may be free'd with the free_callback method, or it will be
              free'd when the instance is destroyed.

              **Raises:**
	      a TransferException if the request object does not meet the specifications above
              If ret != GLOBUS_SUCCESS then cbHandle is an error string which is thrown in
              TransferException
              """
        
        ok = 0
        reqType = request.get_request_type()
        if reqType ==  "GET" and request._isClient == 0:
            ok = 1
        elif reqType == "PUT" and request._isClient == 1:
            ok = 1
        elif reqType == "APPEND" and request._isClient == 1:
            ok = 1

        if not ok:
            raise TransferException,  "improper use of send_bytes, the request arguement is not valid"

        ret, cbHandle = gassTransferc.send_bytes(request.get_handle(), buffer,
                                                  len(buffer),eof,
                                                  callback, arg)
        
        if not ret == 0:
            ex = TransferException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

    def receive_bytes(self, request, buffer, wait_for_amount,callback, arg):
        """
              **Arguments:**
              - *request* is a TransRequest object it must be client side if it is used in conjunction with a get request
              must be server side if used in conjunction with a put or append request

              - *buffer* A pyGlobus.util.Buffer.Buffer object.. This hides the
              internal globus_byte_t data buffer.

              - *wait_for_amount* the minimum amount of data to wait to receive before the callback function is called,
              the data is received in the callback function

              - *callback* is a function that is called after enough data has been received

              - *arg* a user supplied argument to the callback function

              **Returns:**

              A SWIG'ized pointer to the callback handle is returned. After the callback
              completes, this may be free'd with the free_callback method, or it will be
              free'd when the instance is destroyed.

              **Raises:**
	      A TransferException if the request object does not meet the specifications above 

              If ret != GLOBUS_SUCCESS then cbHandle is an error string which is thrown in
              TransferException

              """
        
        ok = 0
        reqType = request.get_request_type()
        if reqType ==  "GET" and request._isClient == 1:
            ok = 1
        elif reqType == "PUT" and request._isClient == 0:
            ok = 1
        elif reqType == "APPEND" and request._isClient == 0:
            ok = 1
            
        if not ok:
            errString = "improper use of receive_bytes, the request arguement is not valid"    
            ex = TransferException(errString)
            raise ex
        
        
        ret, cbHandle= gassTransferc.receive_bytes(request.get_handle(), buffer.get_handle(),
                                                  buffer.get_size(), wait_for_amount,
                                                 callback, arg)
        
        if not ret == 0:
            ex = TransferException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return cbHandle

class TransferClient(TransData):

    def __init__(self):

        TransData.__init__(self)
        return

    def __del__(self):
        for cb in self._cbList:
            self.free_callback(cb)
            
        return
    

    def register_get(self, attr, url, callback, arg):
        """
              **Arguments:**
              - *attr* is a RequestAttr object

              *url* is the url that is the file that you want to get
              An example of a url could be http://huey.lbl.gov:3730/etc/group

              *callback* is a function that will be called back after the request to get has been authorized, denied or refered

              *arg* is a user arguement to the callback function

              This function is non-blocking

              **Returns**
              ret is the return value from the underlying globus call if ret == 0 then the call was 'successful' in
              the globus sense

              cbHandle is a SWIG'ized pointer to the callback structure, if the call was successful, otherwise it is an
              error message

              request is a SWIG'ized pointer to the underlying request handle.  The request handle is a client side request handle
              
              **Raises**
              A TransferException is raised if the register_get cannot be made
              """
        
        ret,cbHandle,request = gassTransferc.register_get(attr.get_handle(), url, callback, arg)
        req = TransRequest(request, 1)
        
        if not ret == 0:
            ex = TransferException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return req
        
    
    def get(self, attr, url):

        """
              **Arguments:**
              - *attr* is a RequestAttr object

              *url* is the url that is the file that you want to get
              An example of a url could be http://huey.lbl.gov:3730/etc/group

              This function is blocking

              **Returns**
              ret is the return value from the underlying globus call if ret == 0 then the call was 'successful' in
              the globus sense

              request is a SWIG'ized pointer to the underlying request handle. The request handle is a client side request handle
              if ret != 0 then request is an error string and is printed out
              
              **Raises**
              A TransferException is raised if the get cannot be made
              """

        ret, request = gassTransferc.get(attr.get_handle(), url)
        
        if not ret == 0:
            ex = TransferException(request)
            raise ex

        req = TransRequest(request, 1)
        return req
        
    def register_put(self,attr,  url, callback, arg, length=0):
        """
              **Arguments:**
              - *attr* is a RequestAttr object
              
               *url* is the url that is the file that you want to put
              An example of a url could be http://huey.lbl.gov:3730/etc/group

              *length* is the length of the file that you wish to put
              The default 0, will do the put with the length arguement GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
              
              This function is blocking
    
              *callback* is a function that will be called back after the request to get has been authorized, denied or refered

              *arg* is a user arguement to the callback function

              This function is non-blocking

              **Returns**
              ret is the return value from the underlying globus call if ret == 0 then the call was 'successful' in
              the globus sense

              cbHandle is a SWIG'ized pointer to the callback structure, if the call was successful, otherwise it is an
              error message

              request is a SWIG'ized pointer to the underlying request handle. The request handle is a client side request handle

              **Raises**
              A TransferException is raised if the register_put cannot be made
              """

        ret, cbHandle, request = gassTransferc.register_put(attr.get_handle(), length, url, callback, arg)
        req = TransRequest(request, 1)
        if not ret == 0:
            ex = TransferException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return req

    def put(self, attr, url, length=0):
        """
              **Arguments:**
              - *attr* is a RequestAttr object
        
              *url* is the url that is the file that you want to put
              An example of a url could be http://huey.lbl.gov:3730/etc/group

              *length* is the length of the file that you wish to put
              The default 0, will do the put with the length arguement GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
              
              This function is blocking

              **Returns**
              ret is the return value from the underlying globus call if ret == 0 then the call was 'successful' in
              the globus sense

              request is a SWIG'ized pointer to the underlying request handle. The request handle is a client side request handle
        
              **Raises**
	      A TransferException is raised if the put cannot be made
              """
            
        ret, request = gassTransferc.put(attr.get_handle(),url, length)
        req = TransRequest(request, 1)
        if not ret == 0:
             ex = TransferException(request)
             raise ex
        return req
     
    def register_append(self, attr, url, callback, arg, length=0):
        """
              **Arguments:**
              - *attr*  is a RequestAttr object
              
              *url* is the url that is the file that you want to append to
              An example of a url could be http://huey.lbl.gov:3730/etc/group

              *callback* is a function that will be called back after the request to get has been authorized, denied or refered

              *arg* is a user arguement to the callback function

              *length* is the length of the file that you wish to append
              The default 0, will do the put with the length arguement GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
              
              This function is non-blocking

              **Returns**
              ret is the return value from the underlying globus call if ret == 0 then the call was 'successful' in
              the globus sense

              cbHandle is a SWIG'ized pointer to the callback structure, if the call was successful, otherwise it is an
              error message

              request is a SWIG'ized pointer to the underlying request handle. The request handle is a client side request handle

              **Raises**
              A TransferException is raised if the register_append cannot be made
              """

        ret, cbHandle, request = gassTransferc.register_append(attr.get_handle(), length, url, callback, arg)
        req = TransRequest(request, 1)
        if not ret == 0:
            ex = TransferException(cbHandle)
            raise ex
        self._cbList.append(cbHandle)
        return req

    def append(self,attr,url,length=0):
        """
               **Arguments:**
               - *attr*  is a RequestAttr object
        
               - *length* is the length of the file that you wish to append
               The default 0, will do the put with the length arguement GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN
               
               - *url* is the url that is the file that you want to append to
               An example of a url could be http://huey.lbl.gov:3730/etc/group


               This function is blocking

               **Returns**
               ret is the return value from the underlying globus call if ret == 0 then the call was 'successful' in
               the globus sense

               cbHandle is a SWIG'ized pointer to the callback structure, if the call was successful, otherwise it is an
               error message

               request is a SWIG'ized pointer to the underlying request handle. The request handle is a client side request handle

               **Raises**
               A TransferException is raised if the register_append cannot be made
               """
        ret, request = gassTransferc.append(attr.get_handle(),url,length)
        req = TransRequest(request, 1)
        if not ret == 0:
            ex = TransferException(ret)
            raise ex
        return req

    


class TransferServer(TransData):
    """
        __init__ creates a TransListener object which is the listener associated with the TransferServer

        """
    
    def __init__(self,scheme):
        """
              *Args*

              *scheme* is a string of the protocol that will be used
              """
        TransData.__init__(self)
        self._cond = 0
        self._attr = ListenerAttr(scheme)
        self._listener = TransListener(self._attr,scheme)
        return

    def __del__(self):
        
        for cb in self._cbList:
            self.free_callback(cb)

        return
         

    def register_listen(self, callback, arg):
        """
              Has the server listen for incoming client request

              Does not block

              **Arguements**

              *callback* the function that will be called as soon a a client request is registered

              *arg* a user supplied arguement to the callback function

              **Returns**
        
              A SWIG'ized pointer to the callback handle is returned. After the callback
              completes, this may be free'd with the free_callback method, or it will be
              free'd when the instance is destroyed.

              **Raises**

              A GassTransferServerException is Raised if the server is unable to listen for client requests
              """

        ret,cbHandle, listener_handle = gassTransferc.register_listen(self._listener.get_handle(), callback, arg)
        
        if not ret == 0:
            ex =  GassTransferServerException(cbHandle)
            raise ex
        
        self._cbList.append(cbHandle)

    def register_accept(self,attr,callback, arg):
        """
              Has the server accept an incoming client request
              This function should be called after the callback for register_listen has been called back
              Does not block

              **Arguements**

              - *attr* a ListenerAttr object
         
              - *callback* the function that will be called as soon a a client request is registered

              - *arg* a user supplied arguement to the callback function

              **Returns**
        
              A TransRequest object is returned the request object is a server_side request

              **Raises**

              A GassTransferServerException is Raised if the server is unable to accept this client request
              """

        ret, cbHandle, request = gassTransferc.register_accept(attr.get_handle(),
                                                               self._listener.get_handle(), callback, arg)
         
        if not ret == 0:
            ex =  TransferServerException(cbHandle)
            raise ex

        self._cbList.append(cbHandle)
        req = TransRequest(request,0)
        return req

    
    def authorize(self, request, length=0):
        """
              After the server has accepted a request it can authorize the request if it wants the transfer to proceed

              **Arguements**

              - *request* a server side request returned from the register_accept call

              - *length* the number of bytes that the request is authorized to make
              the default 0, will use GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN

              **Returns**
              returns a TransRequest object which will be needed for a send_bytes call
        
              **Raises**
	      A TransferServerException if the request cannnot be authorized

              """
        ret, request_handle = gassTransferc.authorize(request.get_handle(), length)
        if not ret == 0:
            ex =  TransferServerException(request_handle)
            raise ex

        return request

    def deny(self, request, reason, message):

        ret, request = gassTransferc.deny(request, reason, message)
        if not ret == 0:
            ex = TransferServerException(request)
            raise ex

        req = TransRequest(request, 0)
        return req

    
    def get_base_url(self):
        """Calls the listener's get base url function"""
        return self._listener.get_base_url()

    
