############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sha, base64, time
from Ft.Xml import XPath

import ZSI
from ZSI.utility import ElementProxy
from ZSI.wstools.Namespaces import DSIG, OASIS

from pyGridWare.security import VerifyError
from pyGridWare.security.gss.GssSecureConversation import GssSecurityContext
from pyGridWare.security.gss.GssSecureMessage import GSIAuthentication
from pyGridWare.utility.dom import  DomletteElementProxy

from pyGridWare.generated.types import GetNSContainer
import pyGridWare.generated.types.SecureConversation_services_types as AUTH

_xmldsig_class = GetNSContainer(AUTH,"http://www.w3.org/2000/09/xmldsig#")
_wssecurity_secext_class = GetNSContainer(AUTH, "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd")



class Holder: 
    pass

class Signature:
    typecode = _xmldsig_class.Signature_Dec()


class BST(str):
    typecode = AUTH.docs_oasis_open_org_wss_2004_01_oasis_200401_wss_wssecurity_secext_1_0_xsd.BinarySecurityToken_Dec()


class TimeStamp(ZSI.TC.Struct):
    attribute_typecode_dict = {(OASIS.UTILITY,"Id"): ZSI.TC.String() }
    def __init__(self):
        pname = (OASIS.UTILITY,"TimeStamp")
        tclist = [ZSI.TCtimes.gDateTime(pname=(OASIS.UTILITY,"Created"),aname="_Created",minOccurs=1,maxOccurs=1,nillable=0),ZSI.TCtimes.gDateTime(pname=(OASIS.UTILITY,"Expires"),aname="_Expires",minOccurs=1,maxOccurs=1,nillable=0)]
        ZSI.TCcompound.Struct.__init__(self, self.__class__, tclist,pname)


class KeyInfo:
    typecode =_xmldsig_class.KeyInfo_Dec


class SecurityReference:
    typecode = AUTH.docs_oasis_open_org_wss_2004_01_oasis_200401_wss_wssecurity_secext_1_0_xsd.SecurityTokenReference_Dec()


class Reference:
    typecode = AUTH.docs_oasis_open_org_wss_2004_01_oasis_200401_wss_wssecurity_secext_1_0_xsd.Reference_Dec()
    

class GssSignatureHandler:
    """
<wsse:Security soapenv:actor="" soapenv:mustUnderstand="0" xmlns:wsse="http://schemas.xmlsoap.org/ws/2002/04/secext" xmlns:wsu="http://schemas.xmlsoap.org/ws/2002/04/utility">
  <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#">

    <ds:SignedInfo>
      <ds:CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"/>
      <ds:SignatureMethod Algorithm="http://www.globus.org/2002/04/xmlenc#gssapi-sign"/>
      <ds:Reference URI="#digestSource">
        <ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"/>
        <ds:DigestValue>l1flSEDSfkc0THuJP6mV0GsHWwU=</ds:DigestValue>
      </ds:Reference>
    </ds:SignedInfo>

    <ds:SignatureValue>AAAAAAAAAAIAAAM/O2xmRW20i5bhOdmpdDvkEiDHvag=</ds:SignatureValue>

    <ds:KeyInfo>
      <ds:KeyName>00000000-0098-df1f-00000000-000000d009d5</ds:KeyName>
    </ds:KeyInfo>

  </ds:Signature>
</wsse:Security>
    """
    def __init__(self):
        self._can_algo = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
        self._sig_algo = "http://www.globus.org/2002/04/xmlenc#gssapi-sign"
        self._dig_algo = "http://www.w3.org/2000/09/xmldsig#sha1"
        self._contextID = None
        self._expire_time = 300

    def setContextID(self, KeyName):
        self._contextID =  KeyName

    def getContextID(self):
        return self._contextID

    def verify(self, ps):
        """
        The algorithm to verify:

        Get and canonicalize the signedInfo node

        Get the value of the SignatureValue node

        base64 decode the Signature Value

        pass the canonicalized signed info node and the base64 decoded signature to
        verify mic
        """
        
        element = DomletteElementProxy(None, message=ps.header)
        canExp = XPath.Compile('//ds:CanonicalizationMethod')
        nodes = element.evaluate(expression=canExp, processorNss={'ds':DSIG.BASE})
        if len(nodes) is not 0:
            node = nodes[0]
            algo = node.getAttributeValue(None, 'Algorithm')
            if algo != self._can_algo:
                raise VerifyError, \
                    'Unsupported Canonicalization algorithm(%s), Supporting [%s,]' \
                    %(algo, self._can_algo)
            
        idExp = XPath.Compile('//*[@wsu:Id]')
        nodes = element.evaluate(expression=idExp, processorNss={'wsu':OASIS.UTILITY})
        if len(nodes) == 0:
            return

        node = nodes[0]
        cxml = node.canonicalize()
        digestValue = sha.sha(cxml).digest()
        digestExp = XPath.Compile('//ds:DigestValue')
        nodes = element.evaluate(expression=digestExp, processorNss={'ds':DSIG.BASE})
        receivedDigestValue = nodes[0].getValue()
        receivedDigestValueStripped = receivedDigestValue.strip()
        digestValue64 = base64.encodestring(digestValue)
        digestValueStripped = digestValue64.strip()
        if digestValueStripped != receivedDigestValueStripped:
            raise VerifyError, 'DigestValues do not match'

        siExp = XPath.Compile('//ds:SignedInfo')
        nodes = element.evaluate(expression=siExp, processorNss={'ds':DSIG.BASE})
        signedInfo = nodes[0]
        signedInfoStr = signedInfo.canonicalize()
        sigExp = XPath.Compile('//ds:SignatureValue')
        nodes = element.evaluate(expression=sigExp,processorNss={'ds':DSIG.BASE})
        node = nodes[0]
        signatureStr = node.getValue()

        keyExp = XPath.Compile('//ds:KeyName')
        nodes = element.evaluate(expression=keyExp,processorNss={'ds':DSIG.BASE})
        node = nodes[0]

        self._contextID = node.getValue()
        home = GssSecurityContext.getResourceHome()
        secContext = home.get(self._contextID)

        verified = secContext.verifyMIC(signedInfoStr, base64.decodestring(signatureStr))
        if not verified:
            raise VerifyError, 'Signatures do not match'

        
    def sign(self, sw):
        """
        Sign the message whether we are using GssSecureConversation
        or GssSecureMesage
        """

        context_id = self._contextID
        useSecureMessage = False
        idUriList = []
        if context_id is None:
            #Use GSISecureMessage
            useSecureMessage = True
            auth = GSIAuthentication()
            pkiPath = auth.getPkiPath()
            binarySecurityToken = BST(base64.encodestring(pkiPath))
            #move this GssSecureMessage
            tokenRefString = self.getTokenRefString()
            binarySecurityToken._attrs = {(OASIS.UTILITY,'Id'): "CertId-" +tokenRefString,
                        'ValueType':"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1",
                        'EncodingType':"http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"}
            binarySecurityToken.typecode.attribute_typecode_dict.update({(OASIS.UTILITY,'Id'):ZSI.TC.String(),
                                                                         'ValueType': ZSI.TC.String(),
                                                                         'EncodingType':ZSI.TC.String()})
            timeStamp = TimeStamp()
            timeStamp._Created = time.gmtime()
            expires = time.time() + self._expire_time
            timeStamp._Expires = time.gmtime(expires)                                                       
            timeStamp._attrs = {(OASIS.UTILITY,"Id"): "timestamp"}
            idUriList.append("#timestamp")
    
        typecode = AUTH.docs_oasis_open_org_wss_2004_01_oasis_200401_wss_wssecurity_secext_1_0_xsd.Security_Dec()
        pyobj = Holder()
        pyobj._any = []
        if useSecureMessage:
            pyobj._any.append(timeStamp)
            pyobj._any.append(binarySecurityToken)   

        
        sig = Signature()
        sig._SignedInfo = Holder()
        sig._SignedInfo._CanonicalizationMethod = Holder()
        sig._SignedInfo._CanonicalizationMethod._attrs = {'Algorithm':self._can_algo}
        sig._SignedInfo._SignatureMethod = Holder()
        if useSecureMessage:
            sig._SignedInfo._SignatureMethod._attrs = {'Algorithm':"http://www.w3.org/2000/09/xmldsig#rsa-sha1"}
        else:
            sig._SignedInfo._SignatureMethod._attrs = {'Algorithm':self._sig_algo}
        sig._SignedInfo._Reference = []
        sig._SignatureValue = ''
        
        if useSecureMessage:
            sig._KeyInfo = KeyInfo()
            secRef = SecurityReference()
            secRef._any = Reference()
            secRef._any._attrs = {}
            secRef._any._attrs['URI'] = "#CertId-" + tokenRefString
            sig._KeyInfo._any = secRef
        else:
            home = GssSecurityContext.getResourceHome()
            secContext = home.get(context_id)
            sig._KeyInfo = Holder()
            sig._KeyInfo._KeyName = context_id
        
        idExp = XPath.Compile('//*[@wsu:Id]')
        nodes = sw.dom.evaluate(expression=idExp, processorNss={'wsu':OASIS.UTILITY})
        #Calculate the digests in the body
        for node in nodes:
            URI = '#%s' %node.getAttributeValue(OASIS.UTILITY, 'Id')
            ref = Holder()
            ref._attrs = {'URI': URI}
            ref._DigestMethod = Holder()
            ref._DigestMethod._attrs = {'Algorithm':self._dig_algo}
            ref._DigestValue = self.getDigestValue(sw, URI)
            sig._SignedInfo._Reference.append(ref)

        #These haven't been serialized yet, the id's to take
        #digests of in the security header
        for id in idUriList:
            ref = Holder()
            ref._attrs = {'URI': id}
            ref._DigestMethod = Holder()
            ref._DigestMethod._attrs = {'Algorithm':self._dig_algo}
            ref._DigestValue = ''
            sig._SignedInfo._Reference.append(ref)

        pyobj._any.append(sig)
        header_pyobjs = {(typecode.nspname,typecode.pname):pyobj}
        sw.serialize_header(pyobj, typecode)
        
        for URI in idUriList:
            digestValue = self.getDigestValue(sw, URI)
            digestIdExp = XPath.Compile("//ds:DigestValue")
            digestNodes = sw.dom.evaluate(expression=digestIdExp,processorNss={'ds':DSIG.BASE})
        
            for digNode in digestNodes:
                #hack, need to make sure that we get the right digestValue node
                #This will probably fail if we have to take digests of more
                #than one node in the SecurityHeader
                if digNode.getValue() == '\n':
                    digNode.createAppendTextNode(base64.encodestring(digestValue))
                
            
        siExp = XPath.Compile('//ds:SignedInfo')
        nodes = sw.dom.evaluate(expression=siExp, processorNss={'ds':DSIG.BASE})
        signedInfo = nodes[0]
        signedInfoStr = signedInfo.canonicalize()

        sigExp = XPath.Compile('//ds:SignatureValue')
        nodes = sw.dom.evaluate(expression=sigExp, processorNss={'ds':DSIG.BASE})
        signatureValueNode = nodes[0]
        if useSecureMessage:
            hashedSignedInfoStr = sha.sha(signedInfoStr).digest()
            signatureValue = base64.encodestring(auth.sign(hashedSignedInfoStr))
        else:
            signatureValue = base64.encodestring(secContext.getMIC(signedInfoStr))
        signatureValueNode.createAppendTextNode(signatureValue)

    def getDigestValue(self, sw, uri):
        '''
<ns1:createCounter xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/" xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ZSI="http://www.zolera.com/schemas/ZSI/" xmlns:ns1="http://counter.com" xmlns:ns2="http://schemas.xmlsoap.org/ws/2002/04/utility" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ns2:Id="digestSource"></ns1:createCounter>
        '''
        digExpString = '//*[@wsu:Id="' + uri[1:] +'"]'
        digExp = XPath.Compile(digExpString)
        nodes = sw.dom.evaluate(expression=digExp, processorNss={'wsu':OASIS.UTILITY})
        node = nodes[0]
        cxml = node.canonicalize()
        digestValue = sha.sha(cxml).digest()
        return digestValue


    def getTokenRefString(self):
        """
        Return a number to identify a BinarySecurityToken
        """
        return '1234565'

