############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

import time, urlparse
from ZSI import _seqtypes, EvaluateException, WSActionException
from ZSI.TC import _get_global_element_declaration, _get_type_definition, \
     _has_type_definition, AnyElement, AnyType
from ZSI.wstools.Namespaces import WSA
from ZSI.wsdl2python import MessageContainer

class Address:
    '''WS-Address
    '''
    def __init__(self, addressTo=None, wsAddressURI=None, action=None):
        self.wsAddressURI = wsAddressURI
        self.anonymousURI = None
        self._addressTo = addressTo
        self._messageID = None
        self._action = action
        self._endPointReference = None
        self._replyTo = None
        self._relatesTo = None
        self.setUp()

    def setUp(self):
        if (self.wsAddressURI is None and _has_type_definition(WSA.ADDRESS2004, 'EndpointReferenceType') is True)\
           or (self.wsAddressURI is WSA.ADDRESS2004):
            self.wsAddressURI,self.anonymousURI = WSA.ADDRESS2004,WSA.ANONYMOUS2004
        elif (self.wsAddressURI is None and _has_type_definitions(WSA.ADDRESS, 'EndpointReferenceType') is True)\
           or (self.wsAddressURI is WSA.ADDRESS):
            self.wsAddressURI,self.anonymousURI = WSA.ADDRESS,WSA.ANONYMOUS
        else:
            raise EvaluateException,\
                'enabling wsAddressing requires the inclusion of that namespace'
        self._replyTo = self.anonymousURI

    def _checkAction(self, action, value):
        '''WS-Address Action
        action -- Action value expecting.
        value  -- Action value server returned.
        '''
        if action is None:
            raise WSActionException, 'User did not specify WSAddress Action value to expect'
        if not value:
            raise WSActionException, 'missing WSAddress Action, expecting %s' %action
        if value != action:
            raise WSActionException, 'wrong WSAddress Action(%s), expecting %s'%(value,action)

    def _checkFrom(self, pyobj):
        '''WS-Address From, 
        XXX currently not checking the hostname, not forwarding messages.
        pyobj  -- From server returned.
        '''
        value = pyobj._Address
        if value != self._addressTo:
            scheme,netloc,path,query,fragment = urlparse.urlsplit(value)
            hostport = netloc.split(':')
            schemeF,netlocF,pathF,queryF,fragmentF = urlparse.urlsplit(self._addressTo)
            if scheme==schemeF and path==pathF and query==queryF and fragment==fragmentF:
                netloc = netloc.split(':') + ['80']
                netlocF = netlocF.split(':') + ['80']
                if netloc[1] == netlocF[1]:
                    return

            raise WSActionException, 'wrong WS-Address From(%s), expecting %s'%(value,self._addressTo)

    def _checkRelatesTo(self, value):
        '''WS-Address From
        value  -- From server returned.
        '''
        if value != self._messageID:
            raise WSActionException, 'wrong WS-Address RelatesTo(%s), expecting %s'%(value,self._messageID)
        
    def _checkReplyTo(self, value):
        '''WS-Address From
        value  -- From server returned in wsa:To
        '''
        if value != self._replyTo:
            raise WSActionException, 'wrong WS-Address ReplyTo(%s), expecting %s'%(value,self._replyTo)

    def setAction(self, action):
        self._action = action

    def getAction(self):
        return self._action

    def getRelatesTo(self):
        return self._relatesTo

    def getMessageID(self):
        return self._messageID

    def getWSAddressTypeCodes(self, **kw):
        '''kw -- namespaceURI keys with sequence of element names.
        '''
        typecodes = []
        try:
            for nsuri,elements in kw.items():
                for el in elements:
                    pyclass = _get_global_element_declaration(nsuri, el)
                    typecodes.append(pyclass())
            else:
                pass
        except EvaluateException, ex:
            raise EvaluateException, \
                'To use ws-addressing register typecodes for namespace(%s)' %self.wsAddressURI
        return typecodes

    def checkResponse(self, ps, action):
        ''' 
        ps -- ParsedSoap
        action -- ws-action for response
        '''
        namespaceURI = self.wsAddressURI
        d = {namespaceURI:("MessageID","Action","To","From","RelatesTo")}
        typecodes = self.getWSAddressTypeCodes(**d)
        pyobjs = ps.ParseHeaderElements(typecodes)

        self.action = action
        action = pyobjs.get((namespaceURI,"Action"))
        self._checkAction(self.action, action)

        From = pyobjs.get((namespaceURI,"From"))
        self._checkFrom(From)

        RelatesTo = pyobjs.get((namespaceURI,"RelatesTo"))
        self._checkRelatesTo(RelatesTo)

        To = pyobjs.get((namespaceURI,"To"))
        if To: self._checkReplyTo(To)

    def setRequest(self, endPointReference, action):
        '''Call For Request
        '''
        self.header_pyobjs = pyobjs = {}
        namespaceURI = self.wsAddressURI
        self.header_typecodes = typecodes = self.getWSAddressTypeCodes(\
            **{namespaceURI:("MessageID", "Action", 'To', 'From')})
        pyobjs[(namespaceURI, "Action")] = action
        pyobjs[(namespaceURI,"MessageID")] = self._messageID = "uuid:%s" %time.time()
        pyobjs[(namespaceURI, "To")] = self._addressTo

        pyobj = MessageContainer()
        #pyobj.setUp(typecode)
        pyobj._Address = self.anonymousURI
        pyobjs[(namespaceURI,"From")] = pyobj

        if endPointReference:
            pyclass = _get_type_definition(namespaceURI ,'EndpointReferenceType')
            if not isinstance(endPointReference, pyclass):
                raise EvaluateException, 'endPointReference %s is not an EndpointReferenceType (%s)'\
                    %(endPointReference.__class__, pyclass)

            pyclass = _get_type_definition(namespaceURI,'ReferencePropertiesType')
            ReferenceProperties = endPointReference.get_variable(pyclass)

            #Assuming maxOccurs=1 
            any = ReferenceProperties.get_variable(AnyElement)
            for k,v in any.items():
                #self.header_pyobjs[k] = v
                pyobjs[k] = v
                if type(v) in _seqtypes:
                    typecodes.append(AnyType(pname=k, maxOccurs=len(v)))
                else:
                    typecodes.append(AnyType(pname=k))


    def setResponse(self, address, action):
        '''Call For Response
        address -- Request Address
        '''
        self.header_typecodes = typecodes = []
        self.header_pyobjs = pyobjs = {}

        self.replyTo = self.anonymousURI
        self._relatesTo = "uuid:%s" %time.time()
        self.From = None

        namespaceURI = self.wsAddressURI
        typecodes = self.getWSAddressTypeCodes(\
            **{namespaceURI:("MessageID", "Action", 'To',)})

        # wsa:Action
        pyobjs[(namespaceURI, "Action")] = self._action

        # wsa:MessageID
        #self.header_pyobjs[(namespaceURI,"MessageID")] = self._relatesTo
        pyobjs[(namespaceURI,"MessageID")] = address.getRelatesTo()

        # wsa:To
        #if not self.header_pyobjs.has_key((namespaceURI, "From")):
        if address.getFrom() is None:
            self.From = 'http://%s:%s%s' %(self.host,self.port,self.url)
            #self.header_pyobjs[(namespaceURI, "To")] = self.From
            pyobjs[(namespaceURI, "To")] = self.From
        else:
            #to = self.header_pyobjs[(namespaceURI, "From")]
            to = address.getFrom()
            #self.header_pyobjs[(namespaceURI, "To")] = to._Address
            pyobjs[(namespaceURI, "To")] = to._Address
            self.From = to._Address

    def setResponseFromWSAddress(self, address, localURL):
        '''Server-side has to set these fields in response.
        address -- Address instance, representing a WS-Address
        '''
        self.From = localURL
        pyobjs = self.header_pyobjs = {}
        namespaceURI = self.wsAddressURI
        d = {namespaceURI:("MessageID", "Action", "From", "RelatesTo")}
        self.header_typecodes = self.getWSAddressTypeCodes(**d)

        pyobjs[(namespaceURI, "Action")] = self._action
        pyobjs[(namespaceURI,"MessageID")] = "uuid:%s" %time.time()
        pyobjs[(namespaceURI,"RelatesTo")] = address.getMessageID()

        typecode = self.header_typecodes[2]
        pyobj = MessageContainer()
        pyobj.setUp(typecode)
        pyobj._Address = self.From
        pyobjs[(namespaceURI,"From")] = pyobj

    def serialize(self, sw, **kw):
        '''
        sw -- SoapWriter instance, add WS-Address header.
        '''
        typecodes, pyobjs_d = self.header_typecodes,self.header_pyobjs
        for typecode in typecodes:
            nspname,pname = typecode.nspname,typecode.pname
            pyobjs = pyobjs_d.get((nspname,pname))
            if pyobjs is not None:
                sw.serialize_header(pyobjs, typecode, **kw)

    def parse(self, ps, **kw):
        '''
        ps -- ParsedSoap instance
        '''
        namespaceURI = self.wsAddressURI
        elements = ("MessageID","Action","To","From","RelatesTo")
        d = {namespaceURI:elements}
        typecodes = self.getWSAddressTypeCodes(**d)
        pyobjs = ps.ParseHeaderElements(typecodes)
        self._messageID = pyobjs[(namespaceURI,elements[0])]
        self._action = pyobjs[(namespaceURI,elements[1])]
        self._addressTo = pyobjs[(namespaceURI,elements[2])]
        self._from = pyobjs[(namespaceURI,elements[3])]
        self._relatesTo = pyobjs[(namespaceURI,elements[4])]


if __name__ == '__main__': print _copyright

