############################################################################
# Monte M. Goode, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################

# contains text container classes for new generation generator

# $Id: containers.py,v 1.1.2.1 2004/10/21 20:27:13 tyger23 Exp $

import re, sys

from utility import StringWriter, NamespaceAliasDict
from ZSI.wstools import XMLSchema
from ZSI.wstools.Namespaces import SCHEMA
from ZSI.typeinterpreter import BaseTypeInterpreter

ID1 = '    '
ID2 = 2*ID1
ID3 = 3*ID1
ID4 = 4*ID1
ID5 = 5*ID1
ID6 = 6*ID1

CONTAINER_DEBUG = True

class ContainerError(Exception):
    pass

class ContainerBase:

    def __init__(self):
        self.content = StringWriter('\n')
        self.setup   = False

        self.ns    = None
        self.nsad  = NamespaceAliasDict()

        if CONTAINER_DEBUG:
            self.write('# Container: <%s>\n' % self.__class__)

    # - string content methods

    def write(self, s):
        self.content.write(s)

    def writeArray(self, a):
        self.content.write('\n'.join(a))

    def _setContent(self):
        """override in subclasses.  formats the content in the desired way."""
        raise NotImplementedError, 'abstract method not implemented'

    def getvalue(self):
        if not self.setup:
            self._setContent()
            self.setup = True
            
        return self.content.getvalue()

    def __str__(self):
        return self.getvalue()

    # - namespace utility methods
    
    def getNSAlias(self):
        if self.ns:
            return self.nsad.getAlias(self.ns)
        else:
            raise ContainerError, 'no self.ns attr defined in %s' \
                  % self.__class__

    def getNSModuleName(self):
        if self.ns:
            return self.nsad.getModuleName(self.ns)
        else:
            raise ContainerError, 'no self.ns attr defined in %s' \
                  % self.__class__


# -- containers for services file components

class ServiceContainerBase(ContainerBase):
    def __init__(self):
        ContainerBase.__init__(self)
    
class ServiceHeaderContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.types = None

    def _setContent(self):
        
        if self.types:
            self.write('from %s import *' % self.types)
                
        imports = [
            '\nimport urlparse, types',
            'from ZSI.TCcompound import Struct',
            'from ZSI import client',
            'import ZSI'
            ]

        self.writeArray(imports)

class ServiceInterfaceContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.serviceName = None
        self.portInfo = []

    def addPortInfo(self, ptName):
        self.portInfo.append(ptName)

    def _setContent(self):

        interface = [
            '# Interface',
            'class %sInterface:' % self.serviceName,
            ]

        for p in self.portInfo:
            pI = [
                '%sdef get%s(self, portAddress=None, **kw):' % (ID1, p),
                '%sraise NonImplementationError, "not implemented"' % ID2,
                ]

            interface += pI

        self.writeArray(interface)


class ServiceLocatorContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.serviceName = None
        self.portInfo = []

    def addPortInfo(self, ptName, bName, sAddress):
        self.portInfo.append( (ptName, bName, sAddress) )

    def _setContent(self):

        locator = [
            '# Locator',
            'class %sLocator(%sInterface):' % (self.serviceName,
                                               self.serviceName),
            ]

        for p in self.portInfo:
            ptName = p[0]
            bName  = p[1]
            sAdd   = p[2]
            pI = [
                '%s%s_address = "%s"' % (ID1, ptName, sAdd),
                '%sdef get%sAddress(self):' % (ID1, ptName),
                '%sreturn %sLocator.%s_address' % (ID2,
                                                   self.serviceName,ptName),
                '%sdef get%s(self, portAddress=None, **kw):' %(ID1, ptName),
                '%sreturn %sSOAP(portAddress or %sLocator.%s_address, **kw)' \
                % (ID2, bName, self.serviceName, ptName),
                ]
            
            locator += pI

        self.writeArray(locator)

class ServiceOperationsClassContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)

        self.rProp = None
        self.wsA   = None
        self.bName = None

        self.operations = []

    def addOperation(self, op):
        self.operations.append(op)

    def _setContent(self):

        if self.wsA:
            ctorArgs = 'endPointReference=None, **kw'
            epr      = 'self.endPointReference = endPointReference'
        else:
            ctorArgs = '**kw'
            epr      = '# no ws-addressing'

        if self.rProp:
            rprop = 'kw["ResourceProperties"] = ("%s","%s")' % (self.rProp[0],
                                                                self.rProp[1])
        else:
            rprop = '# no resource properties'

        methods = [
            '# Methods',
            'class %sSOAP:' % self.bName,
            '%sdef __init__(self, addr, %s):' % (ID1, ctorArgs),
            '%snetloc = (urlparse.urlparse(addr)[1]).split(":") + [80,]' % ID2,
            '%sif not kw.has_key("host"):' % ID2,
            '%skw["host"] = netloc[0]' % ID3,
            '%sif not kw.has_key("port"):' % ID2,
            '%skw["port"] = int(netloc[1])' % ID3,
            '%sif not kw.has_key("url"):' % ID2,
            '%skw["url"] =  urlparse.urlparse(addr)[2]' % ID3,
            '%s%s' % (ID2, rprop),
            '%sself.binding = client.Binding(wsAddressing=%s, **kw)' \
            %(ID2, self.wsA),
            '%s%s' % (ID2,epr),
            ]

        for op in self.operations:
            methods += [ op.getvalue() ]

        self.writeArray(methods)

class ServiceOperationContainer(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)

        self.name = None
        self.wsA  = None
        self.soapaction = None

        self.inputName  = None
        self.outputName = None
        
        self.inputSimpleType  = None
        self.outputSimpleType = None

        self.inputAction  = None
        self.outputAction = None

    def _setContent(self):

        if self.inputSimpleType:
            kwstring = 'kw = {"requestclass": %s}' % self.inputName
            tCheck = 'if not type(request) == %s:' % self.inputSimpleType
        else:
            kwstring = 'kw = {}'
            tCheck = 'if not isinstance(request, %s):' % self.inputName

        if self.outputSimpleType:
            rCheck = 'if not type(response) == %s:' % self.outputSimpleType
        else:
            rCheck = 'if not isinstance(response, %s) and not issubclass(%s, response.__class__):' % (self.outputName, self.outputName)
            
        if self.wsA:
            wsactionIn = 'wsaction = "%s"' % self.inputAction
            wsactionOut = 'wsaction = "%s"' % self.outputAction
            bindArgs = 'wsaction=wsaction, endPointReference=self.endPointReference, **kw)'
            responseArgs = ', wsaction=wsaction'
        else:
            wsactionIn = '# no input wsaction'
            wsactionOut = '# no output wsaction'
            bindArgs = '**kw)'
            responseArgs = ''

        method = [
            '%s# op: %s' % (ID1, self.name),
            '%sdef %s(self, request):' % (ID1, self.name),
            '%s%s' % (ID2, tCheck),
            '%sraise TypeError, "%%s incorrect request type" %% (%s)' \
            % (ID3, 'request.__class__'),
            '%s%s' % (ID2, kwstring),
            '%s%s' % (ID2, wsactionIn),
            '%sself.binding.Send(None, None, request, soapaction="%s", %s' \
            % (ID2, self.soapaction, bindArgs),
            ]

        if self.outputName:
            response = [
                '%s%s' % (ID2, wsactionOut),
                '%sresponse = self.binding.Receive(%s()%s)' \
                % (ID2, self.outputName, responseArgs),
                '%s%s' % (ID2, rCheck),
                '%sraise TypeError, "%%s incorrect response type" %% (%s)' \
                % (ID3, 'response.__class__'),
                '%sreturn response' % ID2,
                ]
            
            method += response

        self.writeArray(method)

class ServiceMessageContainerBase(ServiceContainerBase):
    def __init__(self):
        ServiceContainerBase.__init__(self)
        
        self.pName = None
        self.mName = None

class ServiceLiteralMessageContainer(ServiceMessageContainerBase):

    def _setContent(self):

        nsa = self.getNSAlias()

        message = [
            'class %s(%s.%s_Dec):' % (self.mName, nsa, self.pName),
            '%stypecode = %s.%s_Dec()' %(ID1, nsa, self.pName),
            '%sdef __init__(self, name=None, ns=None):' % ID1,
            '%s%s.%s_Dec.__init__(self, name=None, ns=None)' % (ID2,nsa,
                                                                self.pName),
            ]

        self.writeArray(message)

class ServiceEncodedMessageContainer(ServiceMessageContainerBase):

    def _setContent(self):

        message = [
            '# bogus encoded message',
            ]

        self.writeArray(message)

# -- containers for types file components

class TypesContainerBase(ContainerBase):
    def __init__(self):
        ContainerBase.__init__(self)


class TypesHeaderContainer(TypesContainerBase):
        
    def _setContent(self):
        
        imports = [
            'import ZSI',
            'import ZSI.TCcompound',
            'from ZSI.generate.utility import ElementDeclaration,TypeDefinition',
        ]

        self.writeArray(imports)

class NamespaceClassContainerBase(TypesContainerBase):
    def __init__(self):
        TypesContainerBase.__init__(self)
        

class NamespaceClassHeaderContainer(NamespaceClassContainerBase):

    def _setContent(self):

        head = [
            '#' * 30,
            '# targetNamespace',
            '# %s' % self.ns,
            '#' * 30 + '\n',
            'class %s:' % self.getNSAlias(),
            '%stargetNamespace = "%s"' % (ID1, self.ns)
            ]

        self.writeArray(head)

class NamespaceClassFooterContainer(NamespaceClassContainerBase):

    def _setContent(self):

        foot = [
            '# end class %s (tns: %s)' % (self.getNSAlias(), self.ns),
            ]

        self.writeArray(foot)

# -- containers for acutal typecode components

class TypecodeContainerBase(TypesContainerBase):
    
    def __init__(self):
        TypesContainerBase.__init__(self)
        self.name = None
        self.bti = BaseTypeInterpreter()

        # attrs for model groups and others with elements, tclists, etc...
        self.allOptional = False
        self.mgContent = None
        self.contentFlattened = False
        self.elementAttrs = []
        self.elementsSet = False
        self.tcListElements = []
        self.tcListSet = False

        self.subTypes = []

        # used when processing nested anonymous types
        self.parentClass = None

        # used when processing attribute content

        self.mixed = False
        self.extraFlags = ''

        self.attributeContent = None
        self.attrComponents = []
        self.attrSet = False

    # following methods used to propogate typecode 'components' to
    # the inhereting container classes

    def mangle(self, s):
        # class/variable name illegalities
        if s:
            return re.sub('[-./:?*~*# ]', '_', s)
        else:
            return None


    def getClassName(self):

        if not self.name:
            raise ContainerError, 'self.name not defined!'
        if not hasattr(self.__class__, 'type'):
            raise ContainerError, 'container type not defined!'

        suffix = ''

        if self.__class__.type == 'def':
            suffix = '_Def'
        elif self.__class__.type == 'dec':
            suffix = '_Dec'
        else:
            raise ContainerError, 'unknown container instance type: %s' \
                  % self.__class__.type
        
        return self.mangle( self.name + suffix )

    def nsuriLogic(self):

        if self.parentClass:
            return 'ns = %s.%s.get_nsuri()' % (self.parentClass,
                                               self.getClassName())
        else:
            return 'ns = %s.%s.get_nsuri()' % ( self.getNSAlias(),
                                                self.getClassName() )

    def schemaTag(self):
        return 'schema = "%s"' % self.ns
    
    def typeTag(self):
        return 'type = (schema, "%s")' % self.name
    
    def literalTag(self):
        return 'literal = "%s"' % self.name

    def pname(self):
        return 'kw["pname"] = self.get_pname()'

    def aname(self):
        return 'kw["aname"] = self.get_aname(kw["pname"][1])'

    def getExtraFlags(self):
        if self.mixed:
            self.extraFlags += 'mixed=True, mixed_aname="_text", '

        return self.extraFlags

    def simpleConstructor(self, superclass=None):

        if superclass:
            return '%s.__init__(self, **kw)' % superclass
        else:
            return 'def __init__(self, **kw):'

    def pnameConstructor(self, superclass=None):

        if superclass:
            return '%s.__init__(self, pname, **kw)' % superclass
        else:
            return 'def __init__(self, pname, **kw):'

    # _flattenContent() is a pre-processor called before either the
    # elements or typecode strings are generated.  it flattens out
    # embedded/anonymous model group elements.  _setElements() and
    # _setTypecodeList() skip the model groups and pay attention
    # to the elements contained therein.

    def _flattenContent(self):
        if not self.contentFlattened:
            self.contentFlattened = True
            for c in self.mgContent:
                if c.isChoice() or c.isSequence():
                    self.mgContent += c.content

    # getElements() and _setElements() generate the class variable
    # elements lists in definition classes

    def _setElements(self):

        if not type(self.mgContent) == type(()):
            raise ContainerError, 'mgContent not set!'

        for c in self.mgContent:
            if c.isDeclaration() and c.isElement():
                if c.getAttribute('name'):
                    e = '%sself._%s = None' \
                        % (ID3, self.mangle(c.getAttribute('name')))
                    self.elementAttrs.append(e)
                elif c.isWildCard():
                    e = '%sself._any = None' % (ID3)
                else:
                    raise ContainerError, 'ouch'
            elif c.isReference():
                if c.getAttribute('ref')[1]:
                    e = '%sself._%s = None' \
                        % (ID3,self.mangle(c.getAttribute('ref')[1]))
                    self.elementAttrs.append(e)
                else:
                    raise ContainerError, 'ouch as well'
            elif c.isSequence() or c.isChoice():
                if not self.contentFlattened:
                    raise ContainerError, 'unflattened content found!'
                else:
                    # model group contents already flattened
                    pass
            else:
                print "WHAT AM I? %s" % c

    def getElements(self):
        if not self.elementsSet:
            self._flattenContent()
            self._setElements()
            self.elementsSet = True

        return '\n'.join(self.elementAttrs)

    # getTypecodeList and _setTypecodeList generates the TClist = []
    # strings where appropriate

    def _setTypecodeList(self):
        if not type(self.mgContent) == type(()):
            raise ContainerError, 'mgContent not set!'

        for c in self.mgContent:
            
            tc = TcListComponentContainer()
            
            tc.setOccurs( self._getOccurs(c) )
            
            if c.isDeclaration() and c.isElement():

                if c.isWildCard():
                    # wildcard types
                    if c.getAttribute('name'):
                        tc.name = c.getAttribute('name')
                        tc.klass = 'ZSI.TC.AnyType'
                    else:
                        tc.setStyle('anyElement')
                        
                elif self._elementContainsLocalDef(c):
                    # element containing a local def
                    
                    tc.name = c.getAttribute('name')
                    tc.klass = 'self.__class__.%s_Dec' % tc.name
                    tc.setStyle('ref')

                    # add class subtype for processing
                    self.subTypes.append(c)
                    
                elif c.getAttribute('type'):
                    
                    if self._isXSD(c):
                        # an xsd primitive
                        tc.name = c.getAttribute('name')
                        tpc = self.bti.get_typeclass(c.getAttribute('type')[1],
                                                     c.getAttribute('type')[0])
                        tc.klass = tpc
                        
                    elif self._isRecursiveElement(c):
                        tc.name = c.getAttribute('name')
                        tc.klass = 'ZSI.TC.AnyType'
                        tc.setStyle('namedAny')

                    else:
                        # a definition
                        tc.name = c.getAttribute('name')
                        ns = c.getAttribute('type')[0]
                        tc.klass = '%s.%s_Def' % (self.nsad.getAlias(ns),
                                                  c.getAttribute('type')[1])

                else:
                    print "WHAT WAS THAT DEC?"
                    
            elif c.isReference():
                # element references
                ns = c.getAttribute('ref')[0]
                tc.klass = '%s.%s_Dec' % (self.nsad.getAlias(ns),
                                          c.getAttribute('ref')[1])
                tc.setStyle('ref')

            elif c.isSequence() or c.isChoice():
                if not self.contentFlattened:
                    raise ContainerError, 'unflattened content found!'
                else:
                    # contents already flat
                    continue
            else:
                print 'WHAT WAS THAT? %s' % c

            self.tcListElements.append(tc)

    def getTypecodeList(self):
        if not self.tcListSet:
            self._flattenContent()
            self._setTypecodeList()
            self.tcListSet = True

        list = []
        for e in self.tcListElements:
            list.append(str(e))

        return ', '.join(list)

    # the following _methods() are utility methods used during
    # TCList generation, et al.
    
    def _elementContainsLocalDef(self, c):
        if c.getTypeDefinition('type') == None:
            if hasattr(c, 'content'):
                if isinstance(c.content, XMLSchema.LocalComplexType):
                    return True
                
        return False

    def _isXSD(self, c):
        if c.getAttribute('type'):
            if c.getAttribute('type')[0] in SCHEMA.XSD_LIST:
                return True
            
        return False

    def _isRecursiveElement(self, c):
        if self.name == c.getAttribute('type')[1]:
            if self.ns == c.getAttribute('type')[0]:
                return True

        return False

    def _getOccurs(self, e):
        
        nillable = e.getAttribute('nillable')
        if nillable == 'true':
            nillable = '1'
        else:
            nillable = '0'
            
        maxOccurs = e.getAttribute('maxOccurs')
        if maxOccurs == 'unbounded':
            maxOccurs = '"%s"' %maxOccurs
            
        minOccurs = e.getAttribute('minOccurs')
        
        if self.allOptional is True:
            #JRB Hack
            minOccurs = '0'
            maxOccurs = '"unbounded"'

        return minOccurs,maxOccurs,nillable

    # these methods are used when generating attribute content:

    def _setAttributes(self):

        if self.attributeContent:
            self.attrComponents.append('# attribute handling code')
            if self.mixed:
                self.attrComponents.append('self.mixed = True')
            self.attrComponents.append('self._attrs = {}')
            self.attrComponents.append('self.attribute_typecode_dict.update(%s.%s.attribute_typecode_dict)' % (self.getNSAlias(), self.getClassName()))
        else:
            return

        for a in self.attributeContent:
            if a.isWildCard() and a.isAttribute() and a.isDeclaration():
                self.attrComponents.append('self.attribute_typecode_dict[("%s","anyAttribute")] = ZSI.TC.AnyElement()' % (SCHEMA.XSD3))
            elif a.isDeclaration():
                t = a.getAttribute('type')
                try:
                    tc = self.bti.get_typeclass(t[1], t[0])
                except:
                    raise ContainerError, 'choked on this attr: %s', t
                key = None
                if a.getAttribute('form') == 'qualified':
                    key = '("%s","%s")' % ( a.getTargetNamespace(),
                                            a.getAttribute('name') )
                elif a.getAttribute('form') == 'unqualified':
                    key = '"%s"' % a.getAttribute('name')
                else:
                    raise ContainerError, \
                          'attribute form must be un/qualified %s' \
                          % a.getAttribute('form')
                self.attrComponents.append('self.attribute_typecode_dict[%s] = %s()' % (key, tc))
                
            elif a.isReference():
                if a.isAttributeGroup():
                    # flatten 'em out....
                    for ga in a.getAttributeGroup().getAttributeContent():
                        if not ga.isAttributeGroup():
                            self.attributeContent += (ga,)
                    continue
                ga = a.getAttributeDeclaration()
                tp = ga.getTypeDefinition('type')
                if tp is None:
                    # built in simple type
                    try:
                        namespace,typeName = ga.getAttribute('type')
                    except TypeError, ex:
                        raise ContainerError,\
                              'no type in attribute declaration!'
                    tc = self.bti.get_typeclass(typeName, namespace)
                    key = '("%s","%s")' \
                          % (ga.getTargetNamespace(),ga.getAttribute('name'))
                    self.attrComponents.append('self.attribute_typecode_dict[%s] = %s()' % (key, tc))
                else:
                    typeName = tp.getAttribute('name')
                    namespace = tp.getTargetNamespace()
                    alias = self.nsad.getAlias(namespace)
                    key = '("%s","%s")' \
                          % (ga.getTargetNamespace(),ga.getAttribute('name'))
                    self.attrComponents.append('self.attribute_typecode_dict[%s] = %s.%s_Def()' % (key, alias, typeName))
                
            else:
                raise ContainerError, 'unknown attribute type!'

        self.attrComponents.append('setattr(self, self.attrs_aname, {})')

    def getAttributes(self, indent):

        if not self.attrSet:
            self._setAttributes()
            self.attrSet = True

        attrs = ''
        
        for a in self.attrComponents:
            attrs += '%s%s\n' % (indent, a)

        return attrs

    def getBasesLogic(self, indent):

        retval = ''

        bases = [
            'if %s.%s_Def not in %s.%s.__bases__:' \
            % (self.nsad.getAlias(self.sKlassNS), self.sKlass,
               self.getNSAlias(), self.getClassName()),
            '%sbases = list(%s.%s.__bases__)' \
            % (ID1,self.getNSAlias(),self.getClassName()),
            '%sbases.insert(0, %s.%s_Def)' \
            % (ID1,self.nsad.getAlias(self.sKlassNS), self.sKlass),
            '%s%s.%s.__bases__ = tuple(bases)' \
            % (ID1, self.getNSAlias(), self.getClassName())
            ]

        for b in bases:
            retval += '%s%s\n' % (indent, b)

        return retval

# this class encapsulates a single value in the TClist list.
# it inherits TypecodeContainerBase only to get the mangle() method,
# it does not call the baseclass ctor.

class TcListComponentContainer(TypecodeContainerBase):
    
    def __init__(self):
        self.name = None
        self.klass = None
        
        self.min = None
        self.max = None
        self.nil = None

        self.style = 'standard'

    def setOccurs(self, tup):
        self.min = tup[0]
        self.max = tup[1]
        self.nil = tup[2]

    def setStyle(self, style):
        if style not in ['standard', 'ref', 'anyElement', 'namedAny']:
            raise ContainerError, '"%s" not a valid TC component style!' \
                  % style
        else:
            self.style = style
            
    def _getOccurs(self):
        return 'minOccurs=%s, maxOccurs=%s, nillable=%s' \
               % (self.min, self.max, self.nil)

    def _getvalue(self):
        if self.style == 'standard':
            return '%s(pname=(ns,"%s"), aname="_%s", %s)' \
                   % (self.klass, self.name,
                      self.mangle(self.name), self._getOccurs())
        elif self.style == 'ref':
            return '%s(%s)' % (self.klass, self._getOccurs())
        elif self.style == 'anyElement':
            return 'ZSI.TC.AnyElement(aname="_any", %s)' % self._getOccurs()
        elif self.style == 'namedAny':
            return 'ZSI.TC.AnyElement(aname="_%s", %s)' \
                   % (self.mangle(self.name), self._getOccurs())

    def __str__(self):
        return self._getvalue()
    

# The following methods contain the actual typecode variation containers.
# the formatting ought to be fairly skeletal in each one, relying on
# base class methods to propogate a good deal of the logic components
# to them.

# placeholder for testing
class GenericDefinitionTypecodeContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        type = [
            '%sclass %s(ZSI.TCcompound.Struct):' % (ID1,self.getClassName()),
            '%s# definition' % ID2,
            '%spass' % ID2,
            ]

        self.writeArray(type)

# placeholder for testing
class GenericElementTypecodeContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        type = [
            '%sclass %s:' % (ID1,self.getClassName()),
            '%s# element' % ID2,
            '%spass' % ID2,
            ]

        self.writeArray(type)

class ElementSimpleTypeContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        element = [
            '%sclass %s(%s, ElementDeclaration):' % (ID1, self.getClassName(),
                                                     self.sKlass),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%s%s' % (ID3, self.pname()),
            '%s%s' % (ID3, self.aname()),
            '%s%s' % (ID3, self.simpleConstructor(self.sKlass)),
            ]

        self.writeArray(element)

class ElementLocalDefContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        element = [
            '%sclass %s(ZSI.TCcompound.Struct, ElementDeclaration):' \
            % (ID1,self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%s' % self.getElements(),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            '%s%s' % (ID3,self.pname()),
            '%s%s' % (ID3,self.aname()),
            '%sZSI.TCcompound.Struct.__init__(self, self.__class__, TClist, inorder=0, **kw)' % ID3,
            ]

        self.writeArray(element)

class ElementGlobalDefContainer(TypecodeContainerBase):

    type = 'dec'

    def _setContent(self):

        element = [
            '%sclass %s(ElementDeclaration):' % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.literalTag()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.simpleConstructor()),
            '%s%s' % (ID3,self.pname()),
            '%s%s' % (ID3,self.aname()),
            '%s'   % self.getBasesLogic(ID3),
            '%s%s.%s_Def.__init__(self, **kw)' \
            % (ID3, self.nsad.getAlias(self.sKlassNS), self.sKlass),
            ]

        self.writeArray(element)

class DerivedTypeDefContainer(TypecodeContainerBase):

    type = 'def'

    def __init__(self):
        TypecodeContainerBase.__init__(self)

        self.restriction = False
        self.extension = False

        # this is used in figuring attrs for extensions
        self.extType = None

    def _setContent(self):

        if self.restriction:
            assigner = '='
        else:
            assigner = '+='

        definition = [
            '%sclass %s(TypeDefinition):' % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            '%s'   % self.getBasesLogic(ID3),
            '%s%s.%s_Def.__init__(self, pname, **kw)' \
            % (ID3, self.nsad.getAlias(self.sKlassNS), self.sKlass),
            '%sself.ofwhat %s tuple(TClist)' %(ID3, assigner),
            '%s'   % self.getAttributes(ID3),
            ]

        self.writeArray(definition)

class ModelGroupContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        definition = [
            '%sclass %s(ZSI.TCcompound.Struct, TypeDefinition):'
            % (ID1, self.getClassName()),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s'   % self.getElements(),
            '%s%s' % (ID3, self.nsuriLogic()),
            '%sTClist = [%s]' % (ID3, self.getTypecodeList()),
            '%sZSI.TCcompound.Struct.__init__(self, self.__class__, TClist, pname=pname, inorder=0, %s**kw)' % (ID3, self.getExtraFlags()),
            '%s'   % self.getAttributes(ID3),
            ]

        self.writeArray(definition)

class SimpleTypeContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        definition = [
            '%sclass %s(%s, TypeDefinition):' \
            % (ID1, self.getClassName(), self.sKlass),
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.pnameConstructor(self.sKlass)),
            
            ]

        self.writeArray(definition)

class SimpleContentContainer(TypecodeContainerBase):

    type = 'def'

    def _setContent(self):

        definition = [
            '%sclass %s(%s, TypeDefinition):' \
            % (ID1, self.getClassName(), self.sKlass),
            '%s%s' % (ID2, self.schemaTag()),
            '%s%s' % (ID2, self.typeTag()),
            '%sattribute_typecode_dict = {}' % ID2,
            '%s%s' % (ID2, self.pnameConstructor()),
            '%s%s' % (ID3, self.pnameConstructor(self.sKlass)),
            '%s'   % self.getAttributes(ID3),
            ]

        self.writeArray(definition)

class DerivedSimpleContentContainer(DerivedTypeDefContainer):

    type = 'def'

class UnionOrListContainer(SimpleContentContainer):

    type = 'def'
